use std::{
    cell::UnsafeCell,
    collections::{HashMap, HashSet},
    marker::{PhantomData, PhantomPinned},
    mem::{self, MaybeUninit},
    pin::Pin,
    ptr::{self, NonNull},
    sync::{
        Arc,
        atomic::{AtomicU32, Ordering},
    },
};

use werkzeug::CachePadded;

use werkzeug::ptr::TaggedAtomicPtr;

// A Queue with multiple receivers and multiple producers, where a producer can send a message to one of any of the receivers (any-cast), or one of the receivers (uni-cast).
// After being woken up from waiting on a message, the receiver will look up the index of the message in the queue and return it.

struct QueueInner<T> {
    receivers: HashMap<ReceiverToken, CachePadded<(Slot<T>, bool)>>,
    messages: Vec<T>,
    _phantom: std::marker::PhantomData<T>,
}

pub struct Queue<T> {
    inner: UnsafeCell<QueueInner<T>>,
    lock: AtomicU32,
}

unsafe impl<T> Send for Queue<T> {}
unsafe impl<T> Sync for Queue<T> where T: Send {}

enum SlotKey {
    Owned(ReceiverToken),
    Indexed(usize),
}

pub struct Receiver<T> {
    queue: Arc<Queue<T>>,
    lock: Pin<Box<(AtomicU32, PhantomPinned)>>,
}

#[repr(transparent)]
pub struct Sender<T> {
    queue: Arc<Queue<T>>,
}

// TODO: make this a linked list of slots so we can queue multiple messages for
// a single receiver
const SLOT_ALIGN: u8 = core::mem::align_of::<usize>().ilog2() as u8;
struct Slot<T> {
    value: UnsafeCell<MaybeUninit<T>>,
    next_and_state: TaggedAtomicPtr<Self, SLOT_ALIGN>,
    _phantom: PhantomData<Self>,
}

impl<T> Slot<T> {
    fn new() -> Self {
        Self {
            value: UnsafeCell::new(MaybeUninit::uninit()),
            next_and_state: TaggedAtomicPtr::new(ptr::null_mut(), 0), // 0 means empty
            _phantom: PhantomData,
        }
    }

    fn is_set(&self) -> bool {
        self.next_and_state.tag(Ordering::Acquire) == 1
    }

    unsafe fn pop(&self) -> Option<T> {
        NonNull::new(self.next_and_state.ptr(Ordering::Acquire))
            .and_then(|next| {
                // SAFETY: The next slot is a valid pointer to a Slot<T> that was allocated by us.
                unsafe { next.as_ref().pop() }
            })
            .or_else(|| {
                if self
                    .next_and_state
                    .swap_tag(0, Ordering::AcqRel, Ordering::Relaxed)
                    == 1
                {
                    // SAFETY: The value is only initialized when the state is set to 1.
                    Some(unsafe { self.value.as_ref_unchecked().assume_init_read() })
                } else {
                    None
                }
            })
    }

    /// the caller must ensure that they have exclusive access to the slot
    unsafe fn push(&self, value: T) {
        if self.is_set() {
            let next = self.next_ptr();
            unsafe {
                (next.as_ref()).push(value);
            }
        } else {
            // SAFETY: The value is only initialized when the state is set to 1.
            unsafe { self.value.as_mut_unchecked().write(value) };
            self.next_and_state
                .set_tag(1, Ordering::Release, Ordering::Relaxed);
        }
    }

    fn next_ptr(&self) -> NonNull<Slot<T>> {
        if let Some(next) = NonNull::new(self.next_and_state.ptr(Ordering::Acquire)) {
            next.cast()
        } else {
            self.alloc_next()
        }
    }

    fn alloc_next(&self) -> NonNull<Slot<T>> {
        let next = Box::into_raw(Box::new(Slot::new()));

        let next = loop {
            match self.next_and_state.compare_exchange_weak_ptr(
                ptr::null_mut(),
                next,
                Ordering::Release,
                Ordering::Acquire,
            ) {
                Ok(_) => break next,
                Err(other) => {
                    if other.is_null() {
                        eprintln!("What the sigma? Slot::alloc_next: other is null");
                        continue;
                    }
                    // next was allocated under us, so we need to drop the slot we just allocated again.
                    #[cfg(feature = "tracing")]
                    tracing::trace!(
                        "Slot::alloc_next: next was allocated under us, dropping it. ours: {:p}, other: {:p}",
                        next,
                        other
                    );
                    _ = unsafe { Box::from_raw(next) };
                    break other;
                }
            }
        };

        unsafe {
            // SAFETY: The next slot is a valid pointer to a Slot<T> that was allocated by us.
            NonNull::new_unchecked(next)
        }
    }
}

impl<T> Drop for Slot<T> {
    fn drop(&mut self) {
        // drop next chain
        if let Some(next) = NonNull::new(self.next_and_state.swap_ptr(
            ptr::null_mut(),
            Ordering::Release,
            Ordering::Relaxed,
        )) {
            // SAFETY: The next slot is a valid pointer to a Slot<T> that was allocated by us.
            // We drop this in place because idk..
            unsafe {
                next.drop_in_place();
                _ = Box::<mem::ManuallyDrop<Self>>::from_non_null(next.cast());
            }
        }

        // SAFETY: The value is only initialized when the state is set to 1.
        if mem::needs_drop::<T>() && self.next_and_state.tag(Ordering::Acquire) == 1 {
            unsafe { self.value.as_mut_unchecked().assume_init_drop() };
        }
    }
}

// const BLOCK_SIZE: usize = 8;
// struct Block<T> {
//     next: AtomicPtr<Block<T>>,
//     slots: [CachePadded<Slot<T>>; BLOCK_SIZE],
// }

/// A token that can be used to identify a specific receiver in a queue.
#[repr(transparent)]
#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq)]
pub struct ReceiverToken(werkzeug::util::Send<NonNull<u32>>);

impl<T> Queue<T> {
    pub fn new() -> Arc<Self> {
        Arc::new(Self {
            inner: UnsafeCell::new(QueueInner {
                messages: Vec::new(),
                receivers: HashMap::new(),
                _phantom: PhantomData,
            }),
            lock: AtomicU32::new(0),
        })
    }

    pub fn new_sender(self: &Arc<Self>) -> Sender<T> {
        Sender {
            queue: self.clone(),
        }
    }

    pub fn num_receivers(self: &Arc<Self>) -> usize {
        let _guard = self.lock();
        self.inner().receivers.len()
    }

    pub fn as_sender(self: &Arc<Self>) -> &Sender<T> {
        unsafe { mem::transmute::<&Arc<Self>, &Sender<T>>(self) }
    }

    pub fn new_receiver(self: &Arc<Self>) -> Receiver<T> {
        let recv = Receiver {
            queue: self.clone(),
            lock: Box::pin((AtomicU32::new(0), PhantomPinned)),
        };

        // allocate slot for the receiver
        let token = recv.get_token();
        let _guard = recv.queue.lock();
        recv.queue
            .inner()
            .receivers
            .insert(token, CachePadded::new((Slot::new(), false)));

        drop(_guard);
        recv
    }

    fn lock(&self) -> impl Drop {
        unsafe {
            let lock = werkzeug::sync::Lock::from_ptr(&self.lock as *const _ as _);
            lock.lock();
            werkzeug::drop_guard::DropGuard::new(|| lock.unlock())
        }
    }

    fn inner(&self) -> &mut QueueInner<T> {
        // SAFETY: The inner is only accessed while the queue is locked.
        unsafe { &mut *self.inner.get() }
    }
}

impl<T> QueueInner<T> {
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    fn poll(&mut self, token: ReceiverToken) -> Option<T> {
        // check if someone has sent a message to this receiver
        let CachePadded((slot, _)) = self.receivers.get(&token)?;

        unsafe { slot.pop() }.or_else(|| {
            // if the slot is empty, we can check the indexed messages
            #[cfg(feature = "tracing")]
            tracing::trace!("QueueInner::poll: checking open messages");

            self.messages.pop()
        })
    }
}

impl<T> Receiver<T> {
    pub fn get_token(&self) -> ReceiverToken {
        // the token is just the pointer to the lock of this receiver.
        // the lock is pinned, so it's address is stable across calls to `receive`.

        ReceiverToken(werkzeug::util::Send(NonNull::from(&self.lock.0).cast()))
    }
}

impl<T> Drop for Receiver<T> {
    fn drop(&mut self) {
        if mem::needs_drop::<T>() {
            // lock the queue
            let _guard = self.queue.lock();
            let queue = self.queue.inner();

            // remove the receiver from the queue
            _ = queue.receivers.remove(&self.get_token());
        }
    }
}

impl<T: Send> Receiver<T> {
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn recv(&self) -> T {
        let token = self.get_token();

        loop {
            // lock the queue
            let _guard = self.queue.lock();
            let queue = self.queue.inner();

            // check if someone has sent a message to this receiver
            if let Some(t) = queue.poll(token) {
                queue.receivers.get_mut(&token).unwrap().1 = false; // mark the slot as not parked
                return t;
            }

            // there was no message for this receiver, so we need to park it
            queue.receivers.get_mut(&token).unwrap().1 = true; // mark the slot as parked

            // wait for a message to be sent to this receiver
            drop(_guard);
            unsafe {
                let lock = werkzeug::sync::Lock::from_ptr(token.0.into_inner().as_ptr());
                lock.wait();
            }
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn try_recv(&self) -> Option<T> {
        let token = self.get_token();

        // lock the queue
        let _guard = self.queue.lock();
        let queue = self.queue.inner();

        // check if someone has sent a message to this receiver
        queue.poll(token)
    }
}

impl<T: Send> Sender<T> {
    /// Sends a message to one of the receivers in the queue, or makes it
    /// available to any receiver that will park in the future.
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn anycast(&self, value: T) {
        let _guard = self.queue.lock();

        // SAFETY: The queue is locked, so we can safely access the inner queue.
        match unsafe { self.try_anycast_inner(value) } {
            Ok(_) => {}
            Err(value) => {
                #[cfg(feature = "tracing")]
                tracing::trace!(
                    "Queue::anycast: no parked receiver found, adding message to indexed slots"
                );

                // no parked receiver found, so we want to add the message to the indexed slots
                let queue = self.queue.inner();
                queue.messages.push(value);

                // waking up a parked receiver is not necessary here, as any
                // receivers that don't have a free slot are currently waking up.
            }
        }
    }

    pub fn try_anycast(&self, value: T) -> Result<(), T> {
        // lock the queue
        let _guard = self.queue.lock();

        // SAFETY: The queue is locked, so we can safely access the inner queue.
        unsafe { self.try_anycast_inner(value) }
    }

    /// The caller must hold the lock on the queue for the duration of this function.
    unsafe fn try_anycast_inner(&self, value: T) -> Result<(), T> {
        // look for a receiver that is parked
        let queue = self.queue.inner();
        if let Some((token, slot)) =
            queue
                .receivers
                .iter()
                .find_map(|(token, CachePadded((slot, is_parked)))| {
                    // ensure the slot is available
                    if *is_parked && !slot.is_set() {
                        Some((*token, slot))
                    } else {
                        None
                    }
                })
        {
            // we found a receiver that is parked, so we can send the message to it
            unsafe {
                slot.value.as_mut_unchecked().write(value);
                slot.next_and_state
                    .set_tag(1, Ordering::Release, Ordering::Relaxed);
                werkzeug::sync::Lock::from_ptr(token.0.into_inner().as_ptr()).wake_one();
            }

            return Ok(());
        } else {
            return Err(value);
        }
    }

    /// Sends a message to a specific receiver, waking it if it is parked.
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn unicast(&self, value: T, receiver: ReceiverToken) -> Result<(), T> {
        // lock the queue
        let _guard = self.queue.lock();
        let queue = self.queue.inner();

        let Some(CachePadded((slot, is_parked))) = queue.receivers.get_mut(&receiver) else {
            return Err(value);
        };

        unsafe {
            slot.push(value);
        }

        // check if the receiver is parked
        if *is_parked {
            // wake the receiver
            unsafe {
                werkzeug::sync::Lock::from_ptr(receiver.0.into_inner().as_ptr()).wake_one();
            }
        }

        Ok(())
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn broadcast(&self, value: T)
    where
        T: Clone,
    {
        // lock the queue
        let _guard = self.queue.lock();
        let queue = self.queue.inner();

        // send the message to all receivers
        for (token, CachePadded((slot, is_parked))) in queue.receivers.iter() {
            // SAFETY: The slot is owned by this receiver.

            unsafe { slot.push(value.clone()) };

            // check if the receiver is parked
            if *is_parked {
                // wake the receiver
                unsafe {
                    werkzeug::sync::Lock::from_ptr(token.0.into_inner().as_ptr()).wake_one();
                }
            }
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn broadcast_with<F>(&self, mut f: F)
    where
        F: FnMut() -> T,
    {
        // lock the queue
        let _guard = self.queue.lock();
        let queue = self.queue.inner();

        // send the message to all receivers
        for (token, CachePadded((slot, is_parked))) in queue.receivers.iter() {
            // SAFETY: The slot is owned by this receiver.

            unsafe { slot.push(f()) };

            // check if the receiver is parked
            if *is_parked {
                // wake the receiver
                unsafe {
                    werkzeug::sync::Lock::from_ptr(token.0.into_inner().as_ptr()).wake_one();
                }
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_queue() {
        let queue = Queue::<i32>::new();

        let sender = queue.new_sender();
        let receiver1 = queue.new_receiver();
        let receiver2 = queue.new_receiver();

        let token2 = receiver2.get_token();

        sender.anycast(42);

        assert_eq!(receiver1.recv(), 42);

        sender.unicast(100, token2).unwrap();
        assert_eq!(receiver1.try_recv(), None);
        assert_eq!(receiver2.recv(), 100);
    }

    #[test]
    fn queue_broadcast() {
        let queue = Queue::<i32>::new();

        let sender = queue.new_sender();
        let receiver1 = queue.new_receiver();
        let receiver2 = queue.new_receiver();

        sender.broadcast(42);

        assert_eq!(receiver1.recv(), 42);
        assert_eq!(receiver2.recv(), 42);
    }

    #[test]
    fn queue_multiple_messages() {
        let queue = Queue::<i32>::new();

        let sender = queue.new_sender();
        let receiver = queue.new_receiver();

        sender.anycast(1);
        sender.unicast(2, receiver.get_token()).unwrap();

        assert_eq!(receiver.recv(), 2);
        assert_eq!(receiver.recv(), 1);
    }

    #[test]
    fn queue_threaded() {
        #[derive(Debug, Clone, Copy)]
        enum Message {
            Send(i32),
            Exit,
        }

        let queue = Queue::<Message>::new();

        let sender = queue.new_sender();

        let threads = (0..5)
            .map(|_| {
                let queue_clone = queue.clone();
                let receiver = queue_clone.new_receiver();

                std::thread::spawn(move || {
                    loop {
                        match receiver.recv() {
                            Message::Send(value) => {
                                println!("Receiver {:?} Received: {}", receiver.get_token(), value);
                            }
                            Message::Exit => {
                                println!("Exiting thread");
                                break;
                            }
                        }
                    }
                })
            })
            .collect::<Vec<_>>();

        // Send messages to the receivers
        for i in 0..10 {
            sender.anycast(Message::Send(i));
        }

        // Send exit messages to all receivers
        sender.broadcast(Message::Exit);
        for thread in threads {
            thread.join().unwrap();
        }
        println!("All threads have exited.");
    }

    #[test]
    fn drop_slot() {
        // Test that dropping a slot does not cause a double free or panic
        let slot = Slot::<i32>::new();
        unsafe {
            slot.push(42);
            drop(slot);
        }
    }

    #[test]
    fn drop_slot_chain() {
        struct DropCheck<'a>(&'a AtomicU32);
        impl Drop for DropCheck<'_> {
            fn drop(&mut self) {
                self.0.fetch_sub(1, Ordering::SeqCst);
            }
        }

        impl<'a> DropCheck<'a> {
            fn new(counter: &'a AtomicU32) -> Self {
                counter.fetch_add(1, Ordering::SeqCst);
                Self(counter)
            }
        }
        let counter = AtomicU32::new(0);
        let slot = Slot::<DropCheck>::new();
        for _ in 0..10 {
            unsafe {
                slot.push(DropCheck::new(&counter));
            }
        }
        assert_eq!(counter.load(Ordering::SeqCst), 10);
        drop(slot);
        assert_eq!(
            counter.load(Ordering::SeqCst),
            0,
            "All DropCheck instances should have been dropped"
        );
    }
}

// struct AtomicLIFO<T> {
//     cell: AtomicOption<T>,
//     next: AtomicPtr<Self>,
// }

// impl<T> AtomicLIFO<T> {
//     fn new() -> Self {
//         Self {
//             cell: AtomicOption::new(),
//             next: AtomicPtr::new(ptr::null_mut()),
//         }
//     }

//     fn new_with_value(value: T) -> Self {
//         Self {
//             cell: AtomicOption::from_option(Some(value)),
//             next: AtomicPtr::new(ptr::null_mut()),
//         }
//     }

//     /// inserts a value into the chain, either inserting it into the current
//     /// cell, or allocating a new cell to store it in and atomically linking it
//     /// to the chain.
//     fn push(&self, value: T) {
//         match self.cell.swap(Some(value)) {
//             Some(old) => {
//                 // there was previously a value in this cell, so we have to
//                 // allocate a new cell and link it
//                 let next = Box::into_raw(Box::new(Self::new_with_value(old)));

//                 let mut current_next = self.next.load(Ordering::Acquire);
//                 loop {
//                     unsafe {
//                         (&*next).next.store(current_next, Ordering::Relaxed);
//                     }

//                     match self.next.compare_exchange_weak(
//                         current_next,
//                         next,
//                         Ordering::Release,
//                         Ordering::Relaxed,
//                     ) {
//                         Ok(_) => {
//                             // we successfully linked the new cell, so we can return
//                             return;
//                         }
//                         Err(next) => {
//                             // the next pointer was changed, so we need to try again
//                             current_next = next;
//                         }
//                     }
//                 }
//             }
//             None => {}
//         }
//     }

//     fn pop(&self) -> Option<T> {
//         // try to take the value from the current cell
//         // if there is no value, then we are done, because there will also be no next cell
//         let value = self.cell.take()?;

//         // try to atomically swap
//     }

//     fn alloc_next(&self) -> NonNull<Self> {
//         let next = Box::into_raw(Box::new(Self::new()));

//         match self.next.compare_exchange_weak(
//             ptr::null_mut(),
//             next,
//             Ordering::Release,
//             Ordering::Relaxed,
//         ) {
//             Ok(next) => unsafe { NonNull::new_unchecked(next) },
//             Err(other) => {
//                 // next was allocated under us, so we need to drop the slot we just allocated again.
//                 _ = unsafe { Box::from_raw(next) };
//                 unsafe { NonNull::new_unchecked(other) }
//             }
//         }
//     }
// }
